#Credits to: https://github.com/detkov/Convolution-From-Scratch/
import torch
import numpy as np
from typing import List, Tuple, Union


def calc_out_dims(matrix, kernel_side, stride, dilation, padding):
    batch_size, n_channels, n, m = matrix.shape

    # 使用 PyTorch 操作替代 NumPy
    h_out = torch.floor((torch.tensor(n, dtype=torch.float32) + 2 * padding[0] - kernel_side -
                         (kernel_side - 1) * (dilation[0] - 1)) / stride[0]).int() + 1

    w_out = torch.floor((torch.tensor(m, dtype=torch.float32) + 2 * padding[1] - kernel_side -
                         (kernel_side - 1) * (dilation[1] - 1)) / stride[1]).int() + 1

    return h_out.item(), w_out.item(), batch_size, n_channels

def multiple_convs_kan_conv2d(matrix, #but as torch tensors. Kernel side asume q el kernel es cuadrado
             kernels, 
             kernel_side,
             out_channels,
             stride= (1, 1), 
             dilation= (1, 1), 
             padding= (0, 0),
             device= "cuda"
             ) -> torch.Tensor:
    """Makes a 2D convolution with the kernel over matrix using defined stride, dilation and padding along axes.

    Args:
        matrix (batch_size, colors, n, m]): 2D matrix to be convolved.
        kernel  (function]): 2D odd-shaped matrix (e.g. 3x3, 5x5, 13x9, etc.).
        stride (Tuple[int, int], optional): Tuple of the stride along axes. With the `(r, c)` stride we move on `r` pixels along rows and on `c` pixels along columns on each iteration. Defaults to (1, 1).
        dilation (Tuple[int, int], optional): Tuple of the dilation along axes. With the `(r, c)` dilation we distancing adjacent pixels in kernel by `r` along rows and `c` along columns. Defaults to (1, 1).
        padding (Tuple[int, int], optional): Tuple with number of rows and columns to be padded. Defaults to (0, 0).

    Returns:
        np.ndarray: 2D Feature map, i.e. matrix after convolution.
    """
    h_out, w_out,batch_size,n_channels = calc_out_dims(matrix, kernel_side, stride, dilation, padding)
    n_convs = len(kernels)
    matrix_out = torch.zeros((batch_size,out_channels,h_out,w_out)).to(device)#estamos asumiendo que no existe la dimension de rgb
    unfold = torch.nn.Unfold((kernel_side,kernel_side), dilation=dilation, padding=padding, stride=stride)
    conv_groups = unfold(matrix[:,:,:,:]).view(batch_size, n_channels,  kernel_side*kernel_side, h_out*w_out).transpose(2, 3)#reshape((batch_size,n_channels,h_out,w_out))
    #for channel in range(n_channels):
    kern_per_out = len(kernels)//out_channels
    #print(len(kernels),out_channels)
    for c_out in range(out_channels):
        out_channel_accum = torch.zeros((batch_size, h_out, w_out), device=device)

        # Aggregate outputs from each kernel assigned to this output channel
        for k_idx in range(kern_per_out):
            kernel = kernels[c_out * kern_per_out + k_idx]
            conv_result = kernel.conv.forward(conv_groups[:, k_idx, :, :].flatten(0, 1))  # Apply kernel with non-linear function
            out_channel_accum += conv_result.view(batch_size, h_out, w_out)

        matrix_out[:, c_out, :, :] = out_channel_accum  # Store results in output tensor
    
    return matrix_out
def add_padding(matrix: np.ndarray, 
                padding: Tuple[int, int]) -> np.ndarray:
    """Adds padding to the matrix. 

    Args:
        matrix (np.ndarray): Matrix that needs to be padded. Type is List[List[float]] casted to np.ndarray.
        padding (Tuple[int, int]): Tuple with number of rows and columns to be padded. With the `(r, c)` padding we addding `r` rows to the top and bottom and `c` columns to the left and to the right of the matrix

    Returns:
        np.ndarray: Padded matrix with shape `n + 2 * r, m + 2 * c`.
    """
    n, m = matrix.shape
    r, c = padding
    
    padded_matrix = np.zeros((n + r * 2, m + c * 2))
    padded_matrix[r : n + r, c : m + c] = matrix
    
    return padded_matrix



import argparse
import os

import torch
from sklearn.metrics import classification_report
from torch.utils.data import DataLoader
from torchvision import transforms
from tqdm import tqdm

from yms_class.models.AlexNet import AlexNet
from yms_class.tools.dataset import CustomDataset
from yms_class.tools.plotting import plot_confusion_matrix


def main(args):
    all_predictions = []
    all_labels = []
    output_dir = args.output_dir
    results_file = os.path.join(output_dir, 'test_results.txt')
    result_info = ['accuracy', 'precision', 'recall', 'f1-score']

    column_widths = [8, 9, 6, 8]  # 根据实际需要调整宽度
    device = torch.device('cuda:0' if torch.cuda.is_available() else "cpu")
    print("Using {} device training.".format(device.type))
    data_transform = transforms.Compose([
        transforms.Resize((224, 224)),  # 将图像的大小调整为224x224像素
        transforms.ToTensor(),  # 将图像从PIL.Image格式转换为PyTorch张量格式。
        transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),  # 对图像进行归一化，使其具有指定的均值和标准差。
    ])

    data_path = os.path.join(args.data_dir, 'test')
    batch_size = args.batch_size

    test_dataset = CustomDataset(root_dir=data_path, transform=data_transform)
    test_loader = DataLoader(dataset=test_dataset, batch_size=batch_size, shuffle=False)

    model = AlexNet(num_classes=4)
    weights_path = args.weights_path
    weights_dict = torch.load(weights_path, map_location='cpu', weights_only=True)
    model.load_state_dict(weights_dict)
    model = model.to(device)

    model.eval()
    test_iterator = tqdm(test_loader)
    with torch.no_grad():
        for images, labels in test_iterator:
            images = images.to(device)
            labels = labels.to(device)
            outputs = model(images)

            _, predicted = torch.max(outputs, 1)
            all_predictions.extend(predicted.cpu().numpy())
            all_labels.extend(labels.cpu().numpy())

    plot_confusion_matrix(all_labels=all_labels, all_predictions=all_predictions,
                          classes=test_dataset.classes, path=output_dir, name='test_confusion_matrix.png')
    result = classification_report(y_true=all_labels, y_pred=all_predictions,
                                   target_names=test_dataset.classes, digits=4, output_dict=True)
    print(result)


def parse_args(args=None):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('--data_dir', type=str, default=r'D:\Code\deep-learning-algorithms\KAN\data\工况1')
    parser.add_argument('--output_dir', type=str, default=r'output')
    parser.add_argument('--batch_size', default=20, type=int, metavar='N', help='batch size when training.')
    parser.add_argument('--weights_path', default=r'output/best_model.pth',
                        help='pre-trained path')
    return parser.parse_args(args if args else [])


if __name__ == '__main__':
    opts = parse_args()
    print(opts)
    os.makedirs(opts.output_dir)
    main(opts)



import os
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import Dataset, DataLoader
from torchvision import transforms, models
from PIL import Image
import torch.nn.functional as F
import matplotlib.pyplot as plt
from sklearn.metrics import confusion_matrix
import numpy as np
# import seaborn as sns

train_accuracies = []
val_accuracies = []
train_losses = []
val_losses = []
# 自定义训练数据集类
class CustomTrainDataset(Dataset):
    def __init__(self, root_dir, transform=None):
        self.root_dir = root_dir
        # 将数据增强和转换组合在一起
        self.transform = transform or transforms.Compose([
            transforms.RandomResizedCrop(224),  # 随机裁剪到224x224
            transforms.RandomHorizontalFlip(),   # 随机水平翻转
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
        ])
        # 获取root_dir目录下的所有文件夹，并将其存入self.classes列表中
        self.classes = [d for d in os.listdir(root_dir) if os.path.isdir(os.path.join(root_dir, d))]
        # 将self.classes中的文件夹名称作为键，索引值作为值，存入self.class_to_idx字典中
        self.class_to_idx = {cls_name: i for i, cls_name in enumerate(self.classes)}
        # 构建图像路径和标签的列表
        self.images = []
        for cls_name in self.classes:
            dir_path = os.path.join(self.root_dir, cls_name)
            for img_name in os.listdir(dir_path):
                if img_name.endswith('.jpg'):
                    self.images.append((os.path.join(dir_path, img_name), self.class_to_idx[cls_name]))

   # 定义一个名为__len__的方法，用于返回images列表的长度
    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        # 获取图片路径和标签
        image_path, label = self.images[idx]
        # 以RGB模式打开图片
        image = Image.open(image_path).convert('RGB')
        # 如果进行了变换，则进行变换
        if self.transform:
            image = self.transform(image)
        # 返回图片和标签
        return image, label


# 自定义验证数据集类
# 自定义Dataset类
class CustomValidationDataset(Dataset):
    def __init__(self, root_dir, transform=None, class_to_label=None):
        self.root_dir = root_dir
        self.transform = transform
        self.class_to_label = class_to_label if class_to_label is not None else {}
        self.images = [f for f in os.listdir(root_dir) if f.endswith('.bmp')]

        # 如果没有提供class_to_label字典，我们在这里创建它
        if not self.class_to_label:
            self._create_class_to_label_mapping()

    def _create_class_to_label_mapping(self):
        # 假设类别是从0开始编号的连续整数
        classes = sorted(set([filename.split('_')[0] for filename in self.images]))
        self.class_to_label = {cls: i for i, cls in enumerate(classes)}

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        # 获取图片路径
        image_path = os.path.join(self.root_dir, self.images[idx])
        # 打开图片并转换为RGB格式
        image = Image.open(image_path).convert('RGB')

        # 如果有变换，则进行变换

        if self.transform:
            image = self.transform(image)

        # 提取文件名中的类别
        base_filename = os.path.splitext(self.images[idx])[0]
        class_name = base_filename.split('_')[0]
        # 将类别转换为标签
        label = self.class_to_label[class_name]

        return image, label

def plot_confusion_matrix(cm, classes,
                              normalize=False,
                              title='Confusion matrix',
                              cmap=plt.cm.Blues):
        """
        绘制混淆矩阵的函数
        这个函数不修改原始数据，但会返回混淆矩阵。
        """
        if normalize:
            cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
            print("Normalized confusion matrix")
        else:
            print('Confusion matrix, without normalization')

        plt.imshow(cm, interpolation='nearest', cmap=cmap)
        plt.title(title)
        plt.colorbar()
        tick_marks = np.arange(len(classes))
        plt.xticks(tick_marks, classes, rotation=45)
        plt.yticks(tick_marks, classes)

        fmt = '.2f' if normalize else 'd'
        #用于判断变量normalize的值。如果normalize为True，则将格式化字符串.2f赋值给变量fmt；否则，将格式化字符串'd'赋值给变量fmt。
        #其中，.2f表示保留两位小数，'d'表示以十进制形式显示。
        thresh = cm.max() / 2.
        for i, j in np.ndindex(cm.shape):
            plt.text(j, i, format(cm[i, j], fmt),
                     horizontalalignment="center",
                     color="white" if cm[i, j] > thresh else "black")

        plt.tight_layout()
        plt.ylabel('True label')
        plt.xlabel('Predicted label')

    # 数据预处理


transform = transforms.Compose([
    transforms.Resize((224, 224)),#将图像的大小调整为224x224像素
    transforms.ToTensor(),#将图像从PIL.Image格式转换为PyTorch张量格式。
    transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]), #对图像进行归一化，使其具有指定的均值和标准差。
])

# 初始化验证集Dataset
# 定义验证集图片目录
validation_dir = r'D:\Code\data\data\val'  # 替换为你的验证集图片目录
# 创建验证集数据集对象
validation_dataset = CustomValidationDataset(root_dir=validation_dir, transform=transform)

# 创建DataLoader
validation_loader = DataLoader(dataset=validation_dataset, batch_size=16, shuffle=False)

# 数据预处理
transform1 = transforms.Compose([
    transforms.Resize((224, 224)),
    transforms.ToTensor(),
    transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
])

# 训练集数据加载器

# 定义训练集的目录
train_dir = r'D:\Code\data\data\train'
# 创建自定义训练数据集类
train_dataset = CustomTrainDataset(root_dir=train_dir, transform=transform1)
# 创建数据加载器
train_loader = DataLoader(dataset=train_dataset, batch_size=16, shuffle=True)

# 使用resnet18、googlenet、vgg16、alexnet四种模型训练，其中vgg16没有全连接层

model_leibie=input('请输入模型类别：')

if model_leibie=='resnet18':
   # 加载ResNet-18预训练模型
   model = models.resnet18(pretrained=True)

elif model_leibie=='googlenet':
    # 加载googlenet预训练模型
    model = models.googlenet(pretrained=True)

elif model_leibie=='vgg16':
    #加载vgg16模型
    model = models.vgg16(pretrained=True)

elif model_leibie=='alexnet':
    #加载alexnet模型
    model = models.alexnet(pretrained=True)

# 设置设备
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
print(device)
model = model.to(device)

# 损失函数和优化器
criterion = nn.CrossEntropyLoss()  #定义了损失函数为交叉熵损失函数。
optimizer = optim.Adam(model.parameters(), lr=0.001,weight_decay=0.001) #定义了优化器为Adam优化器，学习率为0.001，权重衰减为0.001。

# 定义学习率衰减的调度器
# 使用StepLR学习率调度器，设置学习率下降的步长为1，下降率为0.1
scheduler = optim.lr_scheduler.StepLR(optimizer, step_size=1, gamma=0.1)

# 训练模型
num_epochs = 20
for epoch in range(num_epochs):
    # 训练模型
    model.train()
    running_loss = 0.0
    running_corrects_train = 0
    total_samples_train = 0

    for images, labels in train_loader:
        # 将数据转移到设备
        images, labels = images.to(device), labels.to(device)
        # 梯度清零
        optimizer.zero_grad()
        # 前向传播
        outputs = model(images)
        # 计算损失
        loss = criterion(outputs, labels)
        # 反向传播
        loss.backward()
        optimizer.step()

        running_loss += loss.item() * images.size(0)   # 计算训练损失
        _, preds = torch.max(outputs, 1)    # 计算预测值
        running_corrects_train += torch.sum(preds == labels).item()         # 计算训练准确率
        total_samples_train += labels.size(0) # 计算训练样本总数

    # 计算训练损失和训练准确率
    epoch_loss = running_loss / total_samples_train
    train_accuracy = running_corrects_train / total_samples_train
    print(f'Epoch {epoch + 1}/{num_epochs}, Loss: {epoch_loss:.4f}, Train Accuracy: {train_accuracy:.4f}')

    # 更新学习率
    scheduler.step()

    # 收集所有预测结果和真实标签
    all_preds = []
    all_labels = []

    # 验证模型
    model.eval()
    running_corrects_val = 0
    total_samples_val = 0
    running_val_loss = 0

    with torch.no_grad():
        # 遍历验证集加载器
        for images, labels in validation_loader:
            images, labels = images.to(device), labels.to(device)    # 将图片和标签转移到设备

            outputs = model(images)  # 运行模型
            loss = criterion(outputs, labels)# 计算损失
            running_val_loss += loss.item() * images.size(0) # 累加损失

            _, preds = torch.max(outputs, 1) # 获取预测结果
            running_corrects_val += torch.sum(preds == labels).item() # 累加正确预测的样本数
            total_samples_val += labels.size(0)  # 累加总样本数

            # 将预测结果和真实标签添加到列表
            all_preds.extend(preds.cpu().numpy())
            all_labels.extend(labels.cpu().numpy())

    # 计算验证准确率、验证损失
    val_loss = running_val_loss / total_samples_val
    val_accuracy = (running_corrects_val / total_samples_val)
    print(f'Validation Loss: {val_loss:.4f}, Validation Accuracy: {val_accuracy:.4f}')

    # 然后在训练循环中，每次epoch结束后添加
    train_accuracies.append(train_accuracy)
    val_accuracies.append(val_accuracy)
    # 将验证损失添加到列表
    val_losses.append(val_loss)
    # 将训练损失添加到列表
    train_losses.append(epoch_loss)

# 计算混淆矩阵
cm = confusion_matrix(all_labels, all_preds, labels=np.arange(len(train_dataset.classes)))

# 绘制混淆矩阵
plt.figure()
plot_confusion_matrix(cm, classes=train_dataset.classes, title='Confusion matrix')


# 在所有epoch结束后添加
plt.figure(figsize=(12, 6))
plt.plot(range(1, num_epochs + 1), train_accuracies, label='Training Accuracy')
plt.plot(range(1, num_epochs + 1), val_accuracies, label='Validation Accuracy')
plt.title('Accuracy over epochs')
plt.xlabel('Epochs')
plt.ylabel('Accuracy')
plt.legend()
plt.grid(True)

# 保存图像到指定文件夹

plt.savefig('D:/data/accuracy_plot.jpg')  # 确保使用正确的引号
# 绘制训练和验证损失
plt.figure(figsize=(12, 6))
plt.plot(range(1, num_epochs + 1), train_losses, label='Training Loss')
plt.plot(range(1, num_epochs + 1), val_losses, label='Validation Loss')
plt.title('Loss over epochs')
plt.xlabel('Epochs')
plt.ylabel('Loss')
plt.legend()
plt.grid(True)
plt.show()
# 保存模型
torch.save(model.state_dict(), 'model.pth')

print('Training complete')


import argparse
import os
import sys
from collections import defaultdict

import torch
from scipy.io import savemat
from torchvision.models.feature_extraction import create_feature_extractor
from tqdm import tqdm

from yms_class.tools.dataset import create_dataloaders


def decea_extract_features(encoder, loader, device):
    """从数据集中提取特征"""
    features = defaultdict(list)
    encoder.eval()
    with torch.no_grad():
        for images, labels in tqdm(loader, desc="Extracting features", colour='blue', file=sys.stdout):
            outputs = encoder(images.to(device))
            for label, feat in zip(labels.cpu().numpy(), outputs.cpu().numpy()):
                features[int(label)].append(feat)
    return features


def extract_features(model, loader, device):
    """从数据集中提取特征"""
    all_features = []
    all_labels = []
    model.eval()
    with torch.no_grad():
        for images, labels in tqdm(loader, desc="Extracting feature...", colour='blue', file=sys.stdout):
            outputs = model(images.to(device))
            # 将特征和标签转换为 numpy 数组
            output_features = outputs.cpu().numpy()
            output_labels = labels.cpu().numpy()
            # 将特征和标签添加到对应的列表中
            all_features.extend(output_features)
            all_labels.extend(output_labels)

    return all_features, all_labels


def main(args):
    save_path = args.save_dir
    os.makedirs(save_path, exist_ok=True)

    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    train_loader, val_loader = create_dataloaders(args.data_dir, args.batch_size,
                                                  train_shuffle=False, test=False)

    data_name = os.path.basename(args.data_dir)
    model = torch.load(args.model_path, map_location='cpu', weights_only=False).to(device)

    train_features, train_labels = extract_features(model, train_loader, device)
    val_features, val_labels = extract_features(model, val_loader, device)
    savemat(os.path.join(save_path, f'{data_name}.mat'),
            {'train_features': train_features, 'train_labels': train_labels,
             'val_features': val_features, 'val_labels': val_labels,
             'class': train_loader.dataset.classes})


def parse_args(args=None):
    parser = argparse.ArgumentParser(description='Generate HSA matrix')
    parser.add_argument('--batch_size', type=int, default=100)
    parser.add_argument('--data_dir', default=r'D:\Code\0-data\滚刀磨损\Hobwear\scale_5')
    parser.add_argument('--model_path',
                        default=r'D:\Code\deep-learning-code\classification\encoder.pt')
    parser.add_argument('--save_dir', default=r'D:\Code\deep-learning-code\output\scale_5')
    return parser.parse_args(args if args else [])


if __name__ == '__main__':
    main(parse_args())



import os

import torch
from PIL import Image
from torchvision import transforms, models


def get_class(class_path):
    # 定义一个空字典来存储类别
    classes_dict = []
    # 打开文件
    with open(class_path, 'r', encoding='utf-8') as file:
        # 遍历文件中的每一行
        for line in file:
            # 去除每行末尾的换行符
            category = line.strip()
            # 将类别添加到字典中，值是其自身
            classes_dict.append(category)
    return classes_dict


def main():
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print("Using {} device training.".format(device.type))
    classes_dict = get_class(args.class_path)
    model = models.resnet101(weights=None)
    weights_dict = torch.load(args.model_path, map_location='cpu')
    model.load_state_dict(weights_dict)
    model = model.to(device)

    image_files = os.listdir(args.image_path)

    transform = transforms.Compose([
        transforms.Resize((224, 224)),  # 将图像的大小调整为224x224像素
        transforms.ToTensor(),  # 将图像从PIL.Image格式转换为PyTorch张量格式。
        transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),  # 对图像进行归一化，使其具有指定的均值和标准差。
    ])

    for image_file in image_files:
        image = Image.open(os.path.join(args.image_path, image_file))
        image = transform(image).unsqueeze(dim=0)

        model.eval()
        with torch.no_grad():
            image = image.to(device)
            output = model(image)
            _, predicted = torch.max(output, 1)
            txt = f'{image_file}的预测结果为:{classes_dict[predicted.item()]}'
            with open('output/predict.txt', "a") as f:
                f.write(txt + "\n")
            print(txt)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument('--image_path', default=r'D:\Code\deep-learning-algorithms\KAN\data\val')
    parser.add_argument('--model_path', default=r'output/best_model.pth')
    parser.add_argument('--class_path', default='classes.txt')
    args = parser.parse_args()
    print(args)
    main()



import sys

import matplotlib.pyplot as plt
import numpy as np
import pywt
from tqdm import tqdm
import os
from scipy.io import loadmat
'''
num: 图片保存时防止名字重复，通过末尾数字区分
total: 保存的图片总量
start_num: 从csv表格的第几行开始读取（一般从第二行读取，0代表第二行）
space: 读取间隔（我这里是每1024个采样点作为一个样本）
sampling_period: 采样率（根据数据集实际情况设置，比如数据集采样率为12kHz，则sampling_period = 1.0 / 12000）
totalscal: 小波变换尺度（我这里是256）
wavename: 小波基函数（morl用的比较多，还有很多如：cagu8，cmor1-1等等）
'''


def img_time_freq(data, total, start_num, end_num, space, sampling_period, totalscal, wavename, class_name, save_path):
    bar_format = '{percentage:.1f}%| {n_fmt}/{total_fmt} [{elapsed}<{remaining}, {rate_fmt}{postfix}]'
    for i in tqdm(range(0, total), bar_format=bar_format, file=sys.stdout, colour='green'):
        # 加载数据集
        signal = data[start_num:end_num]
        start_num += space
        end_num += space

        # 计算小波基函数的中心频率fc,然后根据totalscal 计算参数cparam
        # 通过除以np.arange(totalscal, 0, -1) 来生成一系列尺度值，并存储在scales中
        fc = pywt.central_frequency(wavename)
        cparam = 2 * fc * totalscal
        scales = cparam / np.arange(totalscal, 0, -1)

        # 连续小波变换函数
        coefficients, frequencies = pywt.cwt(signal, scales, wavename, sampling_period)

        # 计算变换系数的幅度
        amp = abs(coefficients)

        # 根据采样周期生成时间轴
        t = np.linspace(1, sampling_period, (end_num-start_num), endpoint=False)

        # 绘制时频图
        image_path = os.path.join(save_path, (class_name + '_' + str(i)+'.jpg'))
        plt.figure(figsize=(42 / 100, 42 / 100))
        plt.contourf(t, frequencies, amp, cmap='jet')
        plt.axis('off')  # 去坐标轴
        plt.xticks([])  # 去x轴刻度
        plt.yticks([])  # 去y轴刻度
        # 去白边
        plt.savefig(image_path, bbox_inches='tight', pad_inches=0)
        plt.close()


# def img_time_freq(data, start_num, end_num, space, sampling_period, totalscal, wavename, save_dir):
#     n = data.shape[1]
#     # for i in range(0, n):
#     bar_format = '{percentage:.1f}%| {n_fmt}/{total_fmt} [{elapsed}<{remaining}, {rate_fmt}{postfix}]'
#     for i in tqdm(range(0, n), bar_format=bar_format):
#         signals = data[:, i]
#         total = int(signals.shape[0] / space)
#         start = start_num
#         end = end_num
#         for j in range(0, total):
#             signal = signals[start:end]
#             start += space
#             end += space
#             # 计算小波基函数的中心频率fc,然后根据totalscal 计算参数cparam
#             # 通过除以np.arange(totalscal, 0, -1) 来生成一系列尺度值，并存储在scales中
#             fc = pywt.central_frequency(wavename)
#             cparam = 2 * fc * totalscal
#             scales = cparam / np.arange(totalscal, 0, -1)
#             # 连续小波变换函数
#             coefficients, frequencies = pywt.cwt(signal, scales, wavename, sampling_period)
#             # 计算变换系数的幅度
#             amp = abs(coefficients)
#             # frequencies.max()
#             # 根据采样周期生成时间轴
#             t = np.linspace(1, sampling_period, 1024, endpoint=False)
#             # 绘制时频图
#             plt.figure(figsize=(42 / 100, 42 / 100))
#             plt.contourf(t, frequencies, amp, cmap='jet')
#             plt.axis('off')  # 去坐标轴
#             plt.xticks([])  # 去x轴刻度
#             plt.yticks([])  # 去y轴刻度
#             # image_name = r"D:\Code\0-data\2-滚刀磨损数据集\工况2(2db)"
#             image_name = os.path.join(save_dir, str(i) + '_' + str(j))
#             plt.savefig("{}_resized.jpg".format(image_name.split(".jpg")[0]), bbox_inches='tight', pad_inches=0)
#             plt.close()


def time_freq(data, num, total, start_num, end_num, space, sampling_period, totalscal, wavename, image_path):
    for i in tqdm(range(0, total)):
        # for i in range(0, total):
        # data = data.loc[start_num:end_num, 'data']
        signals = data[start_num:end_num]
        # 计算小波基函数的中心频率fc,然后根据totalscal 计算参数 cparam
        # 通过除以np.arange(totalscal, 0, -1) 来生成一系列尺度值，并存储在scales中
        fc = pywt.central_frequency(wavename)
        cparam = 2 * fc * totalscal
        scales = cparam / np.arange(totalscal, 0, -1)

        # 连续小波变换函数
        coefficients, frequencies = pywt.cwt(signals, scales, wavename, sampling_period)

        # 计算变换系数的幅度
        amp = abs(coefficients)
        # frequencies.max()

        # 根据采样周期生成时间轴
        t = np.linspace(1, sampling_period, 1024, endpoint=False)

        # 绘制时频图
        image_name = os.path.join(image_path, str(num) + '.jpg')
        plt.figure(figsize=(224 / 100, 224 / 100))
        plt.contourf(t, frequencies, amp, cmap='jet')
        plt.axis('off')  # 去坐标轴
        plt.xticks([])  # 去x轴刻度
        plt.yticks([])  # 去y轴刻度
        # 去白边
        plt.savefig(image_name, bbox_inches='tight', pad_inches=0)
        plt.close()
        start_num += space
        end_num += space
        num += 1


def calculate_parameter(folder_path, L=None, M=None, N=None):
    """
    根据已知的两个参数（L, M, N中的两个）计算第三个参数的最大可行值。

    参数:
    folder_path (str): .mat文件目录
    L (int, optional): 窗口长度
    M (int, optional): 数据块数
    N (int, optional): 步长

    返回:
    int: 第三个参数的最大可行值，若无法满足返回-1

    异常:
    ValueError: 参数错误或文件错误时抛出
    """
    # 参数校验
    known_params = [p for p in [L, M, N] if p is not None]
    if len(known_params) != 2:
        raise ValueError("必须且只能提供两个参数")
    if L is not None and L < 0:
        raise ValueError("L必须≥0")
    if M is not None and M < 1:
        raise ValueError("M必须≥1")
    if N is not None and N < 0:
        raise ValueError("N必须≥0")

    # 确定目标参数
    target = None
    if L is None:
        target = 'L'
        M_val, N_val = M, N
    elif M is None:
        target = 'M'
        L_val, N_val = L, N
    else:
        target = 'N'
        L_val, M_val = L, M

    min_value = float('inf')
    valid = True

    for filename in os.listdir(folder_path):
        if not filename.endswith('.mat'):
            continue

        filepath = os.path.join(folder_path, filename)
        try:
            mat = loadmat(filepath)
            signal = mat['DE'].flatten()
            S = len(signal)
        except KeyError:
            raise ValueError(f"文件 {filename} 缺少DE信号")
        except Exception as e:
            raise IOError(f"读取文件 {filename} 失败: {str(e)}")

        try:
            if target == 'N':
                # 已知L和M求N
                if M_val == 1:
                    if S < L_val:
                        valid = False
                        break
                    current = 0
                else:
                    if S < L_val:
                        valid = False
                        break
                    max_step = (S - L_val) // (M_val - 1)
                    current = min(max_step, L_val)
                    if L_val + (M_val - 1) * current > S:
                        valid = False
                        break

            elif target == 'M':
                # 已知L和N求M
                if N_val == 0:
                    if S < L_val:
                        valid = False
                        break
                    current = 1
                else:
                    if L_val < N_val:
                        valid = False
                        break
                    if S < L_val:
                        valid = False
                        break
                    max_m = (S - L_val) // N_val + 1
                    if (max_m - 1) * N_val > S - L_val:
                        max_m -= 1
                    current = max_m if max_m >= 1 else 0

            elif target == 'L':
                # 已知M和N求L
                required_min_signal = M_val * N_val
                if S < required_min_signal:
                    valid = False
                    break
                current = S - (M_val - 1) * N_val
                if current < N_val:
                    valid = False
                    break

            min_value = min(min_value, current)

        except Exception as e:
            valid = False
            break

    if not valid or min_value == float('inf'):
        return -1

    # 最终校验
    if target == 'L' and min_value < N_val:
        return -1
    if target == 'M' and min_value < 1:
        return -1

    return min_value


if __name__ == '__main__':
    data_dir = r'D:\Code\0-data\8-工业大竞赛数据集\mat'
    mat_names = os.listdir(data_dir)
    for mat_name in mat_names:
        class_name = mat_name.split('.')[0]
        mat_path = os.path.join(data_dir, mat_name)
        data = loadmat(mat_path)['DE'].reshape(-1)
        img_time_freq(data, 1000, 0, 1024, 256, 1.0/120000,
                      256, 'morl', class_name, r'D:\Code\0-data\8-工业大竞赛数据集\images')

    # data = loadmat(r'D:\Code\0-data\7-HOB\HOB\1-H1.mat')['DE'].reshape(-1)
    # img_time_freq(data, 1000, 0, 2048, 2048, 1.0/10000,
    #               256, 'morl', '1-H1', r'D:\Code\0-data\0-故障诊断结果输出\HOB\1-H1')
    # M = calculate_parameter(r'D:\Code\0-data\8-工业大竞赛数据集\mat', L=1024, N=256)
    # print(M)



import torch
from torch import nn

from yms_class.convolution.cov import KANConvolution
from yms_class.kan_convolutional.KANConv import KAN_Convolutional_Layer
from yms_class.models.CNN import CNN
from yms_class.tools.tool import count_model_params


class KKAN(nn.Module):
    def __init__(self, num_classes=10):
        super().__init__()
        self.features = nn.Sequential(
            KAN_Convolutional_Layer(in_channels=3, out_channels=4, kernel_size=(3, 3), padding=(1, 1),
                                    grid_size=5),
            KAN_Convolutional_Layer(in_channels=4, out_channels=8, kernel_size=(3, 3), padding=(1, 1),
                                    grid_size=5),
            nn.MaxPool2d(2),
            KAN_Convolutional_Layer(in_channels=8, out_channels=16, kernel_size=(3, 3), padding=(1, 1),
                                    grid_size=5),
            KAN_Convolutional_Layer(in_channels=16, out_channels=32, kernel_size=(3, 3), padding=(1, 1),
                                    grid_size=5),
            nn.MaxPool2d(2),
        )
        self.classifier = nn.Sequential(
            nn.Linear(32 * 8 * 8, 2048),
            nn.ReLU(),
            nn.Linear(2048, 1024),
            nn.ReLU(),
            nn.Linear(1024, num_classes),
        )

    def forward(self, x):
        x = self.features(x)
        x = torch.flatten(x, 1)
        x = self.classifier(x)
        return x

class CCNN(nn.Module):
    def __init__(self, num_classes=10):
        super().__init__()
        self.features = nn.Sequential(
            nn.Conv2d(3, 4, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.Conv2d(4, 8, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Conv2d(8, 16, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.Conv2d(16, 32, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.MaxPool2d(2)
        )
        self.classifier = nn.Sequential(
            nn.Linear(32 * 8 * 8, 2048),
            nn.ReLU(),
            nn.Linear(2048, 1024),
            nn.ReLU(),
            nn.Linear(1024, num_classes),
        )

    def forward(self, x):
        x = self.features(x)
        x = torch.flatten(x, 1)
        x = self.classifier(x)
        return x


class KANCov(nn.Module):
    def __init__(self, num_classes=10):
        super().__init__()
        self.features = nn.Sequential(
            KANConvolution(in_channels=3, out_channels=8, kernel_size=3, padding=1),
            KANConvolution(in_channels=8, out_channels=16, kernel_size=3, padding=1),
            nn.MaxPool2d(2),
            KANConvolution(in_channels=16, out_channels=32, kernel_size=3, padding=1),
            KANConvolution(in_channels=32, out_channels=64, kernel_size=3, padding=1),
            nn.MaxPool2d(2),
        )
        self.classifier = nn.Sequential(
            nn.Linear(64 * 8 * 8, 4096),
            nn.ReLU(),
            nn.Linear(4096, 2048),
            nn.ReLU(),
            nn.Linear(2048, num_classes),
        )

    def forward(self, x):
        x = self.features(x)
        x = torch.flatten(x, 1)
        x = self.classifier(x)
        return x

if __name__ == '__main__':
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    model = KANCov().to(device)
    # kan_num = count_model_params(kan)
    # print(f'KAN卷积的参数量:{kan_num}')
    y = torch.randn(1, 3, 32, 32).to(device)
    z = model(y)
    print(z)



import os
import shutil
import random
from collections import defaultdict


def split_dataset(source_dir, target_root, ratio=(0.8, 0.2), seed=42):
    # 检查源文件夹是否存在
    if not os.path.exists(source_dir):
        print(f"源文件夹 {source_dir} 不存在。")
        return
    # 按前缀分组图片
    prefix_groups = defaultdict(list)
    for filename in os.listdir(source_dir):
        prefix = filename.split('_')[0]
        file_path = os.path.join(source_dir, filename)
        prefix_groups[prefix].append((filename, file_path))

    # 创建目标文件夹
    train_dir = os.path.join(target_root, 'train')
    val_dir = os.path.join(target_root, 'val')
    os.makedirs(train_dir, exist_ok=True)
    os.makedirs(val_dir, exist_ok=True)

    # 对每个分组按比例划分并复制文件
    random.seed(seed)
    for prefix, files in prefix_groups.items():
        total = len(files)
        if total == 0:
            continue
        train_num = int(total * ratio[0])
        val_num = total - train_num

        random.shuffle(files)
        train_files = files[:train_num]
        val_files = files[train_num:]

        def copy_files(file_list, target_dir):
            for filename, src_path in file_list:
                target_file_path = os.path.join(target_dir, filename)
                try:
                    shutil.move(src_path, target_file_path)
                except Exception as e:
                    print(f"复制 {filename} 时出错: {e}")

        copy_files(train_files, train_dir)
        copy_files(val_files, val_dir)
        print(f"前缀 {prefix}：共 {total} 张，划分 train:{len(train_files)} | val:{len(val_files)}")


def copy_images_by_prefix(source_folder, destination_folder, prefix):
    # 检查源文件夹是否存在
    if not os.path.exists(source_folder):
        print(f"源文件夹 {source_folder} 不存在。")
        return
    # 若目标文件夹不存在则创建
    if not os.path.exists(destination_folder):
        os.makedirs(destination_folder)
    # 遍历源文件夹中的所有文件
    for filename in os.listdir(source_folder):
        # 检查文件名是否以指定前缀开头
        if filename.startswith(prefix):
            source_file_path = os.path.join(source_folder, filename)
            destination_file_path = os.path.join(destination_folder, filename)
            try:
                # 复制文件
                shutil.move(source_file_path, destination_file_path)
                print(f"已移动 {filename} 到 {destination_folder}")
            except Exception as e:
                print(f"移动 {filename} 时出错: {e}")


def split_3dataset(source_dir, target_root, ratio=(0.7, 0.2, 0.1), seed=42):
    # 检查源文件夹是否存在
    if not os.path.exists(source_dir):
        print(f"源文件夹 {source_dir} 不存在。")
        return
    # 按前缀分组图片
    prefix_groups = defaultdict(list)
    for filename in os.listdir(source_dir):
        prefix = filename.split('_')[0]
        file_path = os.path.join(source_dir, filename)
        prefix_groups[prefix].append((filename, file_path))

    # 创建目标文件夹
    train_dir = os.path.join(target_root, 'train')
    val_dir = os.path.join(target_root, 'val')
    test_dir = os.path.join(target_root, 'test')
    os.makedirs(train_dir, exist_ok=True)
    os.makedirs(val_dir, exist_ok=True)
    os.makedirs(test_dir, exist_ok=True)

    # 对每个分组按比例划分并复制文件
    random.seed(seed)
    for prefix, files in prefix_groups.items():
        total = len(files)
        if total == 0:
            continue
        train_num = int(total * ratio[0])
        val_num = int(total * ratio[1])
        test_num = total - train_num - val_num

        random.shuffle(files)
        train_files = files[:train_num]
        val_files = files[train_num:train_num + val_num]
        test_files = files[train_num + val_num:]

        def copy_files(file_list, target_dir):
            for filename, src_path in file_list:
                target_file_path = os.path.join(target_dir, filename)
                try:
                    shutil.move(src_path, target_file_path)
                except Exception as e:
                    print(f"复制 {filename} 时出错: {e}")

        copy_files(train_files, train_dir)
        copy_files(val_files, val_dir)
        copy_files(test_files, test_dir)
        print(
            f"前缀 {prefix}：共 {total} 张，划分 train:{len(train_files)} | val:{len(val_files)} | test:{len(test_files)}")


# 使用示例
source_dir = r'D:\Code\0-data\8-工业大竞赛数据集\images'  # 替换为实际的源文件夹路径
target_root = r'D:\Code\0-data\8-工业大竞赛数据集\dataset'  # 替换为实际的目标文件夹路径
split_3dataset(source_dir, target_root)
# A 2-B-007 1-IR-014 2-IR-014 1-OR-007 0-OR-014 1-OR-021
# 使用示例
# source_folder = r'D:\Code\2-ZSL\Zero-Shot-Learning\data\images'  # 替换为实际的源文件夹路径
# destination_folder = r'D:\Code\2-ZSL\Zero-Shot-Learning\data\B\unseen\images'  # 替换为实际的目标文件夹路径
# prefix_list = ['1-IR-007', '2-B-014', '0-OR-021', '2-IR-021', '0-B-014', '1-OR-014']
# for prefix in prefix_list:
#     copy_images_by_prefix(source_folder, destination_folder, prefix)



import torch
from torch import nn


class AlexNet(nn.Module):
    def __init__(self, num_classes: int = 1000, dropout: float = 0.5) -> None:
        super().__init__()
        self.features = nn.Sequential(
            nn.Conv2d(1, 64, kernel_size=11, stride=4, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(64, 192, kernel_size=5, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(192, 384, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(384, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2),
        )
        self.avgpool = nn.AdaptiveAvgPool2d((6, 6))
        self.classifier = nn.Sequential(
            nn.Dropout(p=dropout),
            nn.Linear(256 * 6 * 6, 4096),
            nn.ReLU(inplace=True),
            nn.Dropout(p=dropout),
            nn.Linear(4096, 4096),
            nn.ReLU(inplace=True),
            nn.Linear(4096, num_classes),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.features(x)
        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.classifier(x)
        return x




import collections
import math
import time
from itertools import repeat
from typing import Union, Tuple, TypeVar

import torch
from torch import nn
from torch.nn.parameter import Parameter

from yms_class.convolution.KANLayer import KANLayer

T = TypeVar('T')


def _tuple(n, name="parse"):
    def parse(x):
        if isinstance(x, collections.abc.Iterable):
            return tuple(x)
        return tuple(repeat(x, n))

    parse.__name__ = name
    return parse


_scalar_or_tuple_2_t = Union[T, Tuple[T, T]]
_size_2_t = _scalar_or_tuple_2_t[int]
_pair = _tuple(2, "_pair")

def convolution2d(
        in_channels: int,
        out_channels: int,
        matrix,
        kernels,
        kernel_size,
        stride,
        padding,
        dilation,
        bias,
        unfold
):
    batch_size, in_ch, height, width = matrix.shape
    if in_channels != in_ch:
        raise ValueError(f"Input channel mismatch: parameter in_channels={in_channels}, but input matrix has {in_ch} channels")
    out_height = (height + 2 * padding[0] - dilation[0] * (kernel_size[0] - 1) - 1) // stride[0] + 1
    out_width = (width + 2 * padding[1] - dilation[1] * (kernel_size[1] - 1) - 1) // stride[1] + 1
    # out_height, out_width = calc_out_dims(matrix, kernel_size, stride, padding, dilation)
    matrix_out = matrix.new_zeros((batch_size, out_channels, out_height, out_width))
    # matrix_out = torch.zeros(batch_size, out_channels, out_height, out_width).to(device)
    # unfold = torch.nn.Unfold(kernel_size=kernel_size, dilation=dilation, stride=stride, padding=padding)
    conv_groups = unfold(matrix).view(batch_size, in_channels, kernel_size[0] * kernel_size[1],
                                      out_height * out_width).transpose(2, 3)

    for out_channel in range(out_channels):
        for in_channel in range(in_channels):
            matrix_out[:, out_channel, :, :] += kernels[out_channel][in_channel].forward(
                conv_groups[:, in_channel, :, :].flatten(0, 1)).reshape((batch_size, out_height, out_width))

    # 添加偏置操作：为每个输出通道添加对应的偏置值
    if bias is not None:
        # 将偏置重塑为(1, out_channels, 1, 1)以便广播
        matrix_out += bias.view(1, -1, 1, 1)

    return matrix_out


class KANConvolution(torch.nn.Module):
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: _size_2_t = 3,
                 stride: _size_2_t = 1,
                 padding: _size_2_t = 0,
                 dilation: _size_2_t = 1,
                 num=5, k=3, noise_scale=0.5, scale_base_mu=0.0, scale_base_sigma=1.0,
                 scale_sp=1.0, base_fun=torch.nn.SiLU(), grid_eps=0.02, grid_range=None, sp_trainable=True,
                 sb_trainable=True, sparse_init=False, bias=True
                 ):
        super(KANConvolution, self).__init__()

        if grid_range is None:
            grid_range = [-1, 1]
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = _pair(kernel_size)
        self.stride = _pair(stride)
        self.padding = _pair(padding)
        self.dilation = _pair(dilation)
        self.conv = torch.nn.ModuleList()
        self.unfold = torch.nn.Unfold(kernel_size=kernel_size, dilation=dilation, stride=stride, padding=padding)
        # 根据bias标志创建或忽略偏置参数
        if bias:
            self.bias = Parameter(torch.zeros(self.out_channels))  # 初始化为零
        else:
            self.register_parameter('bias', None)  # 无偏置

        for i in range(self.out_channels):
            self.conv.append(torch.nn.ModuleList())
            for j in range(self.in_channels):
                self.conv[i].append(
                    KANLayer(
                        in_dim=math.prod(self.kernel_size),
                        out_dim=1,
                        num=num,
                        k=k,
                        noise_scale=noise_scale,
                        scale_base_mu=scale_base_mu,
                        scale_base_sigma=scale_base_sigma,
                        scale_sp=scale_sp,
                        base_fun=base_fun,
                        grid_eps=grid_eps,
                        grid_range=grid_range,
                        sp_trainable=sp_trainable,
                        sb_trainable=sb_trainable,
                        sparse_init=sparse_init
                    )
                )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return convolution2d(in_channels=self.in_channels,out_channels=self.out_channels, matrix=x, kernels=self.conv,
                             kernel_size=self.kernel_size, stride=self.stride, padding=self.padding,
                             dilation=self.dilation, bias=self.bias, unfold=self.unfold)

if __name__ == '__main__':
    import torch.nn.functional as F

    device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
    input_size = [1, 3, 32, 32]
    input_map = torch.randn(input_size)
    # kernel_size = [3, 3, 3, 3]
    # kernels = torch.randn(kernel_size)
    # bias = torch.randn(3)
    model = KANConvolution(in_channels=3, out_channels=4, kernel_size=3, stride=1, padding=0, dilation=1)
    t1 = time.time()
    # # for param in model.parameters():
    # #     print(param.device)
    output = model(input_map)
    t2 = time.time()
    # output2 = F.conv2d(input=input_map, weight=kernels, stride=[1, 1], padding=0, bias=bias)
    # # t3 = time.time()
    # print(output.shape)
    # print(output2)
    print(output.shape)
    # print(output2.shape)
    print(f'kan卷积时间：{t2 - t1}')
    # print(f'官方卷积时间：{t3 - t2}')
    # print(f'倍数{(t2 - t1)/(t3 - t2)}')




import os

import numpy as np
from PIL import Image
from torch.utils.data import Dataset, DataLoader, ConcatDataset
from torchvision import transforms


class CustomDataset(Dataset):
    def __init__(self, root_dir, transform, class_to_label=None):
        self.root_dir = root_dir
        self.transform = transform or transforms.Compose([
            transforms.Resize((224, 224)),
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        ])
        self.class_to_label = class_to_label if class_to_label is not None else {}
        self.images = [f for f in os.listdir(root_dir) if f.endswith(('.bmp', '.jpg', '.png'))]

        # 如果没有提供class_to_label字典，我们在这里创建它
        if not self.class_to_label:
            self._create_class_to_label_mapping()
            self.idx_to_labels = {i: cls_name for i, cls_name in enumerate(self.class_to_label)}

    def _create_class_to_label_mapping(self):
        # 假设类别是从0开始编号的连续整数
        self.classes = sorted(set([filename.split('_')[0] for filename in self.images]))
        self.class_to_label = {cls: i for i, cls in enumerate(self.classes)}

    def get_class_to_label(self):
        return self.class_to_label

    def __len__(self):
        return len(self.images)

    def save_label_mapping(self):
        save_dir = os.path.join(os.path.dirname(self.root_dir))
        np.save(os.path.join(save_dir, 'idx_to_labels.npy'), self.idx_to_labels)
        np.save(os.path.join(save_dir, 'classes_to_idx.npy'), self.class_to_label)

    def __getitem__(self, idx):
        # 获取图片路径
        image_path = os.path.join(self.root_dir, self.images[idx])
        # 打开图片并转换为RGB格式
        # image = Image.open(image_path).convert('RGB')
        image = Image.open(image_path)
        # 如果有变换，则进行变换
        if self.transform:
            image = self.transform(image)

        # 提取文件名中的类别
        base_filename = os.path.splitext(self.images[idx])[0]
        class_name = base_filename.split('_')[0]
        # 将类别转换为标签
        label = self.class_to_label[class_name]

        return image, label


def create_dataloaders(data_path, batch_size, transform=transforms.ToTensor(), num_workers=0, subset=False,
                       train_shuffle=True):
    # 训练集数据加载器
    train_dir = os.path.join(data_path, 'train')
    train_dataset = CustomDataset(root_dir=train_dir, transform=transform)
    # 初始化验证集Dataset
    validation_dir = os.path.join(data_path, 'val')  # 替换为你的验证集图片目录
    validation_dataset = CustomDataset(root_dir=validation_dir, transform=transform)
    if not subset:
        train_loader = DataLoader(dataset=train_dataset, batch_size=batch_size, shuffle=train_shuffle,
                                  num_workers=num_workers)
        val_loader = DataLoader(dataset=validation_dataset, batch_size=batch_size, shuffle=False)
        return train_loader, val_loader
    else:
        dataset = ConcatDataset([train_dataset, validation_dataset])
        dataloader = DataLoader(dataset=dataset, batch_size=batch_size, shuffle=False)
        return dataloader

# if __name__ == '__main__':
    # data = CustomDataset(root_dir=r'D:\Code\0-data\3-故障诊断数据集\defect\train', transform=transforms.Compose([]))
    # data.save_label_mapping()



import os
from typing import Union, List

import numpy as np
from matplotlib import pyplot as plt, rcParams
from sklearn.manifold import TSNE
from sklearn.metrics import ConfusionMatrixDisplay, roc_curve, auc, precision_recall_curve
from sklearn.preprocessing import label_binarize


# def plot_confusion_matrix(cm, classes,
#                           save_path='confusion_matrix_D1.png',
#                           normalize=False,
#                           title='Confusion matrix',
#                           cmap=plt.cm.Blues):
#     """
#         绘制混淆矩阵的函数
#         这个函数不修改原始数据，但会返回混淆矩阵。
#         """
#     plt.figure()
#     if normalize:
#         cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
#         print("Normalized confusion matrix")
#     else:
#         print('Confusion matrix, without normalization')
#
#     plt.imshow(cm, interpolation='nearest', cmap=cmap)
#     plt.title(title)
#     plt.colorbar()
#     tick_marks = np.arange(len(classes))
#     plt.xticks(tick_marks, classes, rotation=45)
#     plt.yticks(tick_marks, classes)
#
#     fmt = '.2f' if normalize else 'd'
#     # 用于判断变量normalize的值。如果normalize为True，则将格式化字符串.2f赋值给变量fmt；否则，将格式化字符串'd'赋值给变量fmt。
#     # 其中，.2f表示保留两位小数，'d'表示以十进制形式显示。
#     thresh = cm.max() / 2.
#     for i, j in np.ndindex(cm.shape):
#         plt.text(j, i, format(cm[i, j], fmt),
#                  horizontalalignment="center",
#                  color="white" if cm[i, j] > thresh else "black")
#
#     plt.tight_layout()
#     plt.ylabel('True label')
#     plt.xlabel('Predicted label')
#     plt.savefig(save_path)
#     plt.close()

def plot_confusion_matrix(all_labels,
                          all_predictions,
                          classes,
                          path,
                          name='confusion_matrix.png',
                          normalize=None,
                          cmap=plt.cm.Blues,
                          ):
    ConfusionMatrixDisplay.from_predictions(all_labels,
                                            all_predictions,
                                            display_labels=classes,
                                            cmap=cmap,
                                            normalize=normalize,
                                            xticks_rotation=45
                                            )
    plt.savefig(os.path.join(path, name), dpi=500)
    plt.close()


def plot_multi_class_curves(y_true, y_pred, target_names, save):
    # 将多分类标签转换为二进制标签（One - vs - Rest）
    n_classes = len(set(target_names))
    y_true_bin = label_binarize(y_true, classes=range(n_classes))
    y_pred_bin = label_binarize(y_pred, classes=range(n_classes))

    # 计算每个类别的精确率 - 召回率曲线和 AUC
    precision = dict()
    recall = dict()
    auc_scores = dict()

    for i in range(n_classes):
        precision[i], recall[i], _ = precision_recall_curve(y_true_bin[:, i], y_pred_bin[:, i])
        auc_scores[i] = auc(recall[i], precision[i])

    # 绘制精确率 - 召回率曲线
    plt.figure()
    # 使用更丰富的颜色映射来应对类别数不确定的情况
    cmap = plt.get_cmap('tab10')
    for i in range(n_classes):
        color = cmap(i % 10)  # 循环使用颜色映射中的颜色
        plt.plot(recall[i], precision[i], color=color, lw=2,
                 label=f'{target_names[i]}:{auc_scores[i]:0.4f}')

    plt.xlabel('Recall')
    plt.ylabel('Precision')
    plt.legend(loc="best")
    # 确保保存路径存在
    if not os.path.exists(save):
        os.makedirs(save)
    plt.savefig(os.path.join(save, 'precision_recall_curve.png'), dpi=500)
    plt.close()

    # 计算每个类别的 ROC 曲线和 AUC
    fpr = dict()
    tpr = dict()
    roc_auc = dict()

    for i in range(n_classes):
        fpr[i], tpr[i], _ = roc_curve(y_true_bin[:, i], y_pred_bin[:, i])
        roc_auc[i] = auc(fpr[i], tpr[i])

    # 绘制 ROC 曲线
    plt.figure()
    for i in range(n_classes):
        color = cmap(i % 10)  # 循环使用颜色映射中的颜色
        plt.plot(fpr[i], tpr[i], color=color, lw=2,
                 label=f'{target_names[i]}:{roc_auc[i]:0.4f}')

    plt.plot([0, 1], [0, 1], 'k--', lw=2)
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.0])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.legend(loc="best")
    plt.savefig(os.path.join(save, 'roc_curve.png'), dpi=500)
    plt.close()


def plot_all_metrics(metrics_dict, num_epochs, img_name, save_path, plot_metric=False):
    """
    绘制训练指标曲线

    参数：
        metrics_dict: 包含指标数据的字典，键为指标名称，值为一个列表
        num_epochs: 训练的总轮数
        save_path: 保存图像的路径，默认为 'metrics.png'
    """
    # 检查是否存在 train_losses 和 val_losses
    has_train_loss = 'train_losses' in metrics_dict
    has_val_loss = 'val_losses' in metrics_dict

    # 获取指标名称列表，排除 'epoch'
    metric_names = [key for key in metrics_dict.keys() if key != 'epoch']

    # 如果有 train_losses 和 val_losses，添加一个特殊的键
    if has_train_loss and has_val_loss:
        metric_names.append('train_val_loss')

    # 计算子图的行数和列数，使得 m*n 最接近 metric_names 的数量
    num_metrics = len(metric_names)
    m = int(np.ceil(np.sqrt(num_metrics)))
    n = int(np.ceil(num_metrics / m))

    # 创建子图
    fig, axes = plt.subplots(m, n, figsize=(12 * n, 6 * m))
    axes = axes.flatten()  # 将二维数组的 axes 展平为一维

    # 绘制每个指标的曲线
    for i, name in enumerate(metric_names):
        ax = axes[i]

        if name == 'train_val_loss':
            # 绘制 train_losses 和 val_losses 在同一张图
            train_loss = metrics_dict['train_losses']
            val_loss = metrics_dict['val_losses']
            ax.plot(range(1, num_epochs + 1), train_loss, label='Training Loss')
            ax.plot(range(1, num_epochs + 1), val_loss, label='Validation Loss')
            ax.set_title('Loss over epochs')
        else:
            # 绘制其他指标
            metric = metrics_dict[name]
            ax.plot(range(1, num_epochs + 1), metric, label=f'{name}')
            ax.set_title(f'{name} over epochs')

        ax.set_xlabel('Epochs')
        ax.set_ylabel(f'{name}')
        ax.legend()
        ax.grid(True)

    # 删除多余的子图
    for j in range(i + 1, len(axes)):
        fig.delaxes(axes[j])

    # 调整布局并保存图像
    plt.tight_layout()
    plt.savefig(os.path.join(save_path, f'{img_name}_metrics.png'), dpi=500)
    plt.close()
    if plot_metric:
        # 为每个指标单独绘制一张图
        for name in metric_names:
            plt.figure(figsize=(12, 6))
            if name == 'train_val_loss':
                # 绘制 train_losses 和 val_losses 在同一张图
                train_loss = metrics_dict['train_losses']
                val_loss = metrics_dict['val_losses']
                plt.plot(range(1, num_epochs + 1), train_loss, label='Training Loss')
                plt.plot(range(1, num_epochs + 1), val_loss, label='Validation Loss')
                plt.title('Loss over epochs')
            else:
                # 绘制其他指标
                metric = metrics_dict[name]
                plt.plot(range(1, num_epochs + 1), metric, label=f'{name}')
                plt.title(f'{name} over epochs')

            plt.xlabel('Epochs')
            plt.ylabel(f'{name}')
            plt.legend()
            plt.grid(True)
            plt.savefig(os.path.join(save_path, f'{img_name}_{name}.png'), dpi=500)
            plt.close()


def plot_metrics(metric1, metric2, num_epochs, name, save_path='metrics.png'):
    plt.figure(figsize=(12, 6))
    plt.plot(range(1, num_epochs + 1), metric1, label=f'Training {name}')
    plt.plot(range(1, num_epochs + 1), metric2, label=f'Validation {name}')
    plt.title(f'{name} over epochs')
    plt.xlabel('Epochs')
    plt.ylabel(f'{name}')
    plt.legend()
    plt.grid(True)
    plt.savefig(save_path, dpi=500)
    plt.close()


def plot_single(met, num_epochs, name, save_path='metrics.png'):
    plt.figure(figsize=(12, 6))
    plt.plot(range(1, num_epochs + 1), met)
    plt.title(f'{name} over epochs')
    plt.xlabel('Epochs')
    plt.ylabel(f'{name}')
    plt.grid(True)
    plt.savefig(save_path, dpi=500)
    plt.close()


def plot_data_from_files(file_paths, exclude_headers=None, save_path='metrics.png'):
    """
    Plot data from multiple text files, excluding specified headers, with a fixed 2 rows layout.
    Use folder name as label if file names are consistent, otherwise use file name.

    Args:
    - file_paths (list): A list of file paths to read data from.
    - exclude_headers (list): A list of headers to exclude from plotting.
    """
    # 设置支持中文的字体
    rcParams['font.family'] = 'Microsoft YaHei'  # Windows系统
    # rcParams['font.family'] = 'PingFang SC'  # macOS系统
    rcParams['axes.unicode_minus'] = False  # 正确显示负号
    if exclude_headers is None:
        exclude_headers = ['epoch', 'lr']  # 默认排除的头部

    # 初始化一个列表来存储所有数据
    all_data = []

    # 处理每个文件
    for path in file_paths:
        data = {}
        with open(path, 'r') as file:
            headers = file.readline().strip().split('\t')
            for header in headers:
                data[header] = []
            for line in file:
                values = line.strip().split('\t')
                for header, value in zip(headers, values):
                    data[header].append(float(value))
        all_data.append(data)

    # 提取共同的epochs
    epochs = all_data[0]['epoch']  # 假设所有文件都有相同的epochs

    # 创建子图
    num_metrics = len([m for m in all_data[0].keys() if m not in exclude_headers])
    num_cols = (num_metrics + 1) // 2 + (1 if num_metrics % 2 else 0)  # 计算列数
    fig, axs = plt.subplots(2, num_cols, figsize=(15, 8), constrained_layout=True)
    axs = axs.flatten()  # 展平数组以便更容易迭代

    # 检查文件名是否一致，以确定使用文件名还是文件夹名作为标签
    file_names = [os.path.basename(path) for path in file_paths]
    unique_names = set(file_names)
    if len(unique_names) == 1:
        labels = [os.path.basename(os.path.dirname(path)) for path in file_paths]
    else:
        labels = file_names

    # 绘制每个指标的曲线图
    for i, key in enumerate([m for m in all_data[0].keys() if m not in exclude_headers]):
        for j, data in enumerate(all_data):
            axs[i].plot(epochs, data[key], label=f'{labels[j]} {key}', color=f'C{j}')
        axs[i].set_title(f'{key} over Epochs')
        axs[i].set_xlabel('Epoch')
        axs[i].set_ylabel(key)
        axs[i].legend()

    # 隐藏多余的子图
    for i in range(num_metrics, len(axs)):
        axs[i].axis('off')

    # 显示图表
    plt.savefig(save_path, dpi=500)
    plt.close()


def visualize_features(
        features: Union[List[List[float]], np.ndarray],
        labels: Union[List[int], np.ndarray],
        class_names: List[str],
        n_components: int = 2,
        perplexity: int = 30,
        learning_rate: float = 200.0,
        title: str = "Feature Visualization (t-SNE)",
        save_path: str = None,
        backend: str = "agg",
        markers: Union[str, List[str]] = 'o'
) -> None:
    """
    最终版t-SNE可视化（动态标记适配+警告修复）

    参数:
        markers: 标记形状（自动区分填充/非填充类型）
                 填充标记 (支持白边): 'o','s','^','v','d','p'等
                 非填充标记 (仅颜色): 'x','+','|','_','*'等
                 3D限制: 仅支持 ['o','s','^','v','x','+']（含填充/非填充）
    """
    # 1. 初始化与环境配置
    plt.switch_backend(backend)  # PyCharm兼容后端
    np.random.seed(42)  # 保证可复现性

    # 2. 数据格式校验（严格模式）
    try:
        features = np.asarray(features, dtype=np.float32)
        labels = np.asarray(labels, dtype=np.int32)
    except Exception as e:
        raise TypeError(f"数据格式错误: {str(e)}") from e

    if features.shape[0] != len(labels):
        raise ValueError(f"特征数量({features.shape[0]})与标签数量({len(labels)})不匹配")
    if np.any(labels < 0) or np.max(labels) >= len(class_names):
        raise ValueError(f"标签范围非法: 期望[0,{len(class_names) - 1}], 实际[{np.min(labels)},{np.max(labels)}]")

    # 3. 标记参数校验（含类型/维度校验）
    unique_labels = np.unique(labels)
    n_classes = len(unique_labels)

    if isinstance(markers, str):
        markers = [markers] * n_classes
    elif not isinstance(markers, list) or len(markers) != n_classes:
        raise ValueError(f"markers需为字符串或长度为{class_names}的列表")

    for i, m in enumerate(markers):
        # 3D标记白名单（改为元组） 🌟 修复1
        if n_components == 3 and m not in ('o', 's', '^', 'v', 'x', '+'):  # 元组
            raise ValueError("3D仅支持: o/s/^/v/x/+")
        # 填充标记校验（元组+元组拼接） 🌟 修复2
        if m not in plt.Line2D.filled_markers + ('x', '+', '|', '_', '*'):  # 元组+元组
            raise ValueError(f"未知标记'{m}'，参考Matplotlib文档")

    # 4. T-SNE降维（完整流程含进度显示）
    print(">>> 开始T-SNE降维...")
    tsne = TSNE(
        n_components=n_components,
        perplexity=perplexity,
        learning_rate=learning_rate,
        max_iter=1000,
        random_state=42,
        verbose=1  # 显示训练进度（PyCharm控制台可见）
    )
    reduced = tsne.fit_transform(features)
    print(">>> 降维完成")

    # 5. 颜色与样式准备
    cmap = plt.colormaps['tab10']  # Matplotlib 3.7+推荐
    colors = cmap(unique_labels)  # 按标签索引颜色
    filled_markers = plt.Line2D.filled_markers  # 内置填充标记集合

    # 6. 绘图核心（2D与3D分离实现）
    fig = plt.figure(figsize=(10, 8) if n_components == 2 else (12, 10))

    ax = fig.add_subplot(111)
    if n_components == 2:

        for i, label in enumerate(unique_labels):
            mask = labels == label
            m = markers[i]

            # 动态参数生成（填充/非填充标记区分）
            if m in filled_markers:  # 填充标记（带白边）
                ax.scatter(
                    reduced[mask, 0], reduced[mask, 1],
                    marker=m,
                    facecolors=colors[i],  # 显式指定填充色
                    edgecolors='white',  # 保留白色边缘
                    linewidths=1.2,  # 边缘粗细
                    alpha=0.8,
                    s=60,
                    label=class_names[i]
                )
            else:  # 非填充标记（仅颜色）
                ax.scatter(
                    reduced[mask, 0], reduced[mask, 1],
                    marker=m,
                    color=colors[i],  # 直接控制线条颜色
                    lw=1.5,  # 线条粗细
                    alpha=0.9,
                    s=80,  # 非填充标记适当放大
                    label=class_names[i]
                )

        ax.set(xlabel='t-SNE Dim 1', ylabel='t-SNE Dim 2')

    elif n_components == 3:
        ax = fig.add_subplot(111, projection='3d')
        for i, label in enumerate(unique_labels):
            mask = labels == label
            m = markers[i]

            # 3D特殊处理（填充标记保留白边，非填充仅颜色）
            if m in filled_markers and m in ['o', 's', '^', 'v']:  # 3D填充标记
                ax.scatter(
                    reduced[mask, 0], reduced[mask, 1], reduced[mask, 2],
                    marker=m,
                    facecolors=colors[i],
                    edgecolors='white',
                    linewidths=0.8,  # 3D边缘更细
                    alpha=0.7,
                    s=40,
                    label=class_names[i]
                )
            else:  # 3D非填充标记（x/+）
                ax.scatter(
                    reduced[mask, 0], reduced[mask, 1], reduced[mask, 2],
                    marker=m,
                    color=colors[i],
                    lw=1.2,
                    alpha=0.8,
                    s=50,
                    label=class_names[i]
                )

        ax.set(xlabel='Dim 1', ylabel='Dim 2', zlabel='Dim 3')

    # 7. 通用图表设置
    ax.set_title(title, fontsize=14, pad=20)
    ax.legend(
        title='Classes',
        bbox_to_anchor=(1.05, 1),
        loc='upper left',
        fontsize=11,
        frameon=True,
        framealpha=0.9
    )
    plt.tight_layout(pad=4)  # 防止标签截断

    # 8. 保存/显示（带错误处理）
    if save_path:
        try:
            plt.savefig(
                save_path,
                dpi=500,
                bbox_inches='tight',
                facecolor='white',
                format='png' if '.' not in save_path else None
            )
            print(f"✅ 图片已保存至: {save_path} (尺寸:{fig.get_size_inches()})")
        except Exception as e:
            print(f"❌ 保存失败: {str(e)}")
        finally:
            plt.close(fig)
    elif backend == 'tkagg':
        plt.show()



import torch
from torch import nn


class FeatureExtractor(nn.Module):
    def __init__(self):
        super().__init__()
        self.features = nn.Sequential(
            nn.Conv2d(3, 32, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.Conv2d(32, 64, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Conv2d(64, 128, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.Conv2d(128, 256, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.MaxPool2d(2)
        )
        self.fc = nn.Sequential(
            nn.Linear(256 * 8 * 8, 4096),
            # nn.ReLU(),
            # nn.Linear(4096, 2048),
        )

    def forward(self, x):
        x = self.features(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)
        return x


class CNN(nn.Module):
    def __init__(self, num_classes=10):
        super().__init__()
        self.features = FeatureExtractor()
        self.classifier = nn.Sequential(
            nn.Linear(4096, 2048),
            nn.ReLU(),
            nn.Linear(2048, num_classes)
        )
        # self.fc3 = nn.Linear(2048, num_classes)
        self.relu = nn.ReLU()

    def save(self, path):
        torch.save(self.features, path)

    def forward(self, x):
        x = self.features(x)
        x = self.relu(x)
        x = self.classifier(x)
        return x



import argparse
import os

import torch
from torch.nn import MSELoss
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torchvision import transforms

from yms_class.models.DECAE import DRCAE
from yms_class.tools.dataset import create_dataloaders
from yms_class.tools.plotting import plot_all_metrics
from yms_class.tools.tool import append_to_results_file, initialize_results_file, make_save_dirs
from yms_class.tools.train_eval_utils import train_decae_one_epoch


def main(args):
    save_dir = args.save_dir
    img_dir, model_dir = make_save_dirs(save_dir)

    results_file = os.path.join(save_dir, 'decae_results.txt')
    decae_column_order = ['epoch', 'train_losses', 'val_losses', 'lrs']
    initialize_results_file(results_file, decae_column_order)
    custom_column_widths = {'epoch': 5, 'train_loss': 12, 'val_loss': 10, 'lr': 3}

    device = torch.device('cuda:0' if torch.cuda.is_available() else "cpu")
    print("Using {} device training.".format(device.type))
    if args.num_workers:
        num_workers = max(min(os.cpu_count() - 1, args.batch_size), 0)
    else:
        num_workers = 0
    print(f'use {num_workers} workers')
    data_transform = transforms.Compose([
        transforms.Resize((160, 160)),  # 将图像的大小调整为224x224像素
        transforms.ToTensor(),  # 将图像从PIL.Image格式转换为PyTorch张量格式。
    ])
    train_loader, val_loader = create_dataloaders(args.data_dir, args.batch_size,
                                                  num_workers=num_workers, transform=data_transform)
    metrics = {'train_losses': [], 'val_losses': [], 'lrs': []}

    model = DRCAE(in_channels=1, out_channels=1).to(device)
    optimizer = torch.optim.Adam(model.parameters(), lr=args.lr)
    lr_scheduler = ReduceLROnPlateau(optimizer, 'min', factor=0.1, patience=4, min_lr=1e-9)
    criterion = MSELoss()
    best = 1e8
    for epoch in range(0, args.epochs):
        result = train_decae_one_epoch(model, train_loader, val_loader, device, optimizer, criterion, epoch)
        lr = lr_scheduler.get_last_lr()[0]
        lr_scheduler.step(result['val_loss'])

        metrics['val_losses'].append(result['val_loss'])
        metrics['train_losses'].append(result['train_loss'])
        metrics['lrs'].append(lr)
        result.update({'lr': lr})

        append_to_results_file(results_file, result, decae_column_order,
                               custom_column_widths=custom_column_widths)

        save_file = {
            'epoch': epoch,
            'model_state_dict': model,
            'optimizer_state_dict': optimizer,
            'lr_scheduler_state_dict': lr_scheduler,
        }
        torch.save(save_file, os.path.join(model_dir, 'last_decoder.pt'))
        if result['val_loss'] < best:
            best = result['val_loss']
            model.save(os.path.join(model_dir, 'decoder.pt'))

    plot_all_metrics(metrics, args.epochs, 'decoder', img_dir)
    os.remove(os.path.join(model_dir, 'last_decoder.pt'))


def parse_args(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=str, default=r'D:\Code\0-data\滚刀磨损\Hobwear\scale_5')
    parser.add_argument('--save_dir', type=str, default=r'D:\Code\deep-learning-code\output\scale_5')
    parser.add_argument('--epochs', type=int, default=4)
    parser.add_argument('--lr', type=float, default=1e-3)
    parser.add_argument('--batch_size', type=int, default=32)
    parser.add_argument('--num_workers', type=bool, default=False)
    return parser.parse_args(args if args else [])


if __name__ == '__main__':
    opts = parse_args()
    print(opts)
    main(opts)



import os
import time
from datetime import timedelta

import torch
from torch.nn import CrossEntropyLoss
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torchvision import transforms

from yms_class.models.KKAN import KKAN, KANCov
from yms_class.tools.dataset import create_dataloaders
from yms_class.tools.tool import calculate_metric, append_to_results_file, initialize_results_file, \
    make_save_dirs
from yms_class.tools.train_eval_utils import train_one_epoch, eval_one_epoch


def main(args,run=None):
    # 创建输出文件夹
    save_dir = args.save_dir
    img_dir, model_dir = make_save_dirs(save_dir)

    results_file = os.path.join(save_dir, 'results.txt')
    column_order = ['epoch', 'train_losses', 'val_losses', 'accuracies', 'precisions', 'recalls',
                    'f1-scores', 'lrs']
    initialize_results_file(results_file, column_order)
    custom_column_widths = {'epoch': 5, 'train_loss': 12, 'val_loss': 10, 'accuracy': 10, 'precision': 9, 'recall': 7,
                            'f1-score': 8, 'lr': 3}

    device = torch.device('cuda:0' if torch.cuda.is_available() else "cpu")
    print("Using {} device training.".format(device.type))

    transform = transforms.Compose([
        transforms.Resize((32, 32)),
        transforms.ToTensor(),
        transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
    ])
    train_loader, val_loader = create_dataloaders(args.data_dir, args.batch_size, transform=transform)
    classes = train_loader.dataset.classes
    metrics = {'train_losses': [], 'val_losses': [], 'accuracies': [], 'precisions': [], 'recalls': [], 'f1-scores': [],
               'lrs': []}

    model = KANCov().to(device)
    # model = model.to(device)
    optimizer = torch.optim.Adam(model.parameters(), lr=args.lr)
    lr_scheduler = ReduceLROnPlateau(optimizer, 'min', factor=0.1, patience=5, min_lr=1e-9)
    criterion = CrossEntropyLoss()
    best = -1
    num_epochs = args.epochs

    # 添加早停相关参数
    patience = args.patience  # 从外部参数获取耐心值
    early_stop_counter = 0  # 早停计数器
    best_epoch = 0  # 最佳模型的epoch

    start_time = time.time()
    for epoch in range(0, num_epochs):
        training_lr = lr_scheduler.get_last_lr()[0]
        train_loss, train_accuracy = train_one_epoch(model=model, train_loader=train_loader, device=device,
                                                     optimizer=optimizer, criterion=criterion, epoch=epoch)
        print(f'Epoch {epoch + 1}/{num_epochs}, Loss: {train_loss:.4f}, Train Accuracy: {train_accuracy:.2%},'
              f'lr: {training_lr}')

        result = eval_one_epoch(model=model, val_loader=val_loader,
                                device=device, criterion=criterion, epoch=epoch)
        metric = calculate_metric(result['y_true'], result['y_pred'], classes)
        print(f'val epoch {epoch + 1}, val loss: {result["val_loss"]:.4f}, accuracy: {metric["accuracy"]:.2%}')
        metrics['train_losses'].append(train_loss)
        metrics['val_losses'].append(result['val_loss'])
        metrics['accuracies'].append(metric['accuracy'])
        metrics['precisions'].append(metric['precision'])
        metrics['recalls'].append(metric['recall'])
        metrics['f1-scores'].append(metric['f1-score'])
        metrics['lrs'].append(training_lr)

        metric.update({'epoch': epoch, 'train_loss': train_loss, 'val_loss': result['val_loss'], 'lr': training_lr})
        append_to_results_file(results_file, metric, column_order,
                               custom_column_widths=custom_column_widths)
        if run is not None:
            run.log(metric)

        save_file = {
            'epoch': epoch,
            'model': model,
            'optimizer': optimizer,
            'lr_scheduler': lr_scheduler
        }
        torch.save(save_file, os.path.join(model_dir, 'last_cnn.pt'))

        # 早停逻辑
        current_score = metric['f1-score']
        if current_score > best:
            best = current_score
            best_epoch = epoch
            early_stop_counter = 0  # 重置早停计数器
            torch.save(model, os.path.join(model_dir, 'best_model.pt'))
            print(f'Best model saved at epoch {epoch + 1} with F1-score: {best:.4f}')
        elif patience > 0:  # 仅当patience>0时执行早停计数
            early_stop_counter += 1
            print(f'Early stopping counter: {early_stop_counter}/{patience}')

            # 如果早停计数器达到耐心值，停止训练
            if early_stop_counter >= patience:
                print(f'Early stopping triggered after {epoch + 1} epochs')
                print(f'Best model was saved at epoch {best_epoch + 1} with F1-score: {best:.4f}')
                break

        lr_scheduler.step(result['val_loss'])  # 更新学习率

    total_time = time.time() - start_time
    total_time_str = str(timedelta(seconds=int(total_time)))
    print('Training time {}'.format(total_time_str))
    # plot_all_metrics(metrics, epoch + 1, 'cnn', img_dir)  # 更新为实际训练的epoch数

    os.remove(os.path.join(model_dir, 'last_cnn.pt'))


def parse_args(args=None):
    import argparse
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('--data_dir', default=r'D:\Code\2-ZSL\0-data\data\CRWU\D0')
    parser.add_argument('--save_dir', default=r'kan')
    parser.add_argument('--epochs', default=40, type=int)
    parser.add_argument('--batch_size', default=10, type=int)
    parser.add_argument('--lr', default=1e-3, type=float)
    parser.add_argument('--weight_decay', default=1e-5, type=float)
    parser.add_argument('--patience', default=8, type=int)
    # parser.add_argument('--model_weight', default=r'')

    return parser.parse_args(args if args else [])


if __name__ == '__main__':
    opts = parse_args()
    print(opts)
    main(opts)



# 定义残差块
import torch
from torch import nn


class ResidualBlock(nn.Module):
    def __init__(self, channels):
        super().__init__()
        self.res = nn.Sequential(
            nn.Conv2d(channels, channels, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(channels),
            nn.ReLU(),
            nn.Conv2d(channels, channels, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(channels),
            nn.ReLU(),
        )

    def forward(self, x):
        residual = x
        out = self.res(x)
        return out + residual


# 编码器
class Encoder(nn.Module):
    def __init__(self, in_channels=3,mask_ratio=0.3):
        super().__init__()
        self.mask_ratio = mask_ratio
        # Conv1 + ResidualBlock1
        self.conv1 = nn.Sequential(
            nn.Conv2d(in_channels, 16, 3, stride=2, padding=1),
            nn.BatchNorm2d(16),
            nn.ReLU(),
            ResidualBlock(16)
        )
        # Conv2 + ResidualBlock2
        self.conv2 = nn.Sequential(
            nn.Conv2d(16, 32, 3, stride=2, padding=1),
            nn.BatchNorm2d(32),
            nn.ReLU(),
            ResidualBlock(32)
        )
        # Conv3 + ResidualBlock3
        self.conv3 = nn.Sequential(
            nn.Conv2d(32, 64, 3, stride=2, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(),
            ResidualBlock(64)
        )
        # Conv4 + ResidualBlock4
        self.conv4 = nn.Sequential(
            nn.Conv2d(64, 128, 3, stride=1, padding=1),
            nn.BatchNorm2d(128),
            nn.ReLU(),
            ResidualBlock(128)
        )
        self.flatten = nn.Flatten()
        self.fc = nn.Linear(128 * 20 * 20, 4096)

    def add_noise(self, x):
        if self.training:
            mask = (torch.rand_like(x) > self.mask_ratio).float()
            return x * mask
        return x

    def forward(self, x):
        x = self.add_noise(x)
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.conv4(x)
        x = self.flatten(x)
        x = self.fc(x)
        return x


# 解码器
class Decoder(nn.Module):
    def __init__(self, out_channels=3):
        super().__init__()
        self.fc = nn.Linear(4096, 128*20*20)
        self.unFlatten = nn.Unflatten(1, (128, 20, 20))

        self.deConv = nn.Sequential(
            # DeConv1
            nn.ConvTranspose2d(128, 64, 3, stride=1, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(),
            # DeConv2
            nn.ConvTranspose2d(64, 32, 3, stride=2, padding=1, output_padding=1),
            nn.BatchNorm2d(32),
            nn.ReLU(),
            # DeConv3
            nn.ConvTranspose2d(32, 16, 3, stride=2, padding=1, output_padding=1),
            nn.BatchNorm2d(16),
            nn.ReLU(),
            # DeConv4 (输出层)
            nn.ConvTranspose2d(16, out_channels, 3, stride=2, padding=1, output_padding=1),
        )

    def forward(self, x):
        x = self.fc(x)
        x = self.unFlatten(x)
        x = self.deConv(x)
        return torch.sigmoid(x)  # 假设输出在[0,1]范围
        # return x


class DRCAE(nn.Module):
    def __init__(self, in_channels=3, out_channels=3, mask_ratio=0.3):
        super().__init__()
        self.encoder = Encoder(in_channels, mask_ratio)
        self.decoder = Decoder(out_channels)

    def forward(self, x):
        x = self.encoder(x)
        x = self.decoder(x)
        return x

    def save(self, path):
        torch.save(self.encoder, path)

    def save_decoder(self, path):
        torch.save(self.decoder.state_dict(), path)


if __name__ == '__main__':
    model = DRCAE(in_channels=1, out_channels=1)
    train_input = torch.randn((1, 1, 160, 160))
    output = model(train_input)
    print(output.size())



import sys

import torch
from sklearn.metrics import accuracy_score, classification_report
from tqdm import tqdm


def train_one_epoch(model, train_loader, device, optimizer, criterion, epoch):
    mean_loss = torch.zeros(1).to(device)
    all_predictions = []
    all_labels = []
    model.train()
    train_iterator = tqdm(train_loader, file=sys.stdout, desc=f'the {epoch + 1} epoch is training....', colour='blue')
    for step, (images, labels) in enumerate(train_iterator):
        # 将数据转移到设备
        images, labels = images.to(device), labels.to(device)
        # 梯度清0
        optimizer.zero_grad()
        # 前向传播
        outputs = model(images)
        # 计算损失
        loss = criterion(outputs, labels)
        # 反向传播
        loss.backward()
        # 更新参数
        optimizer.step()

        mean_loss = (mean_loss * step + loss.detach()) / (step + 1)
        _, predicted = torch.max(outputs, 1)
        # 设置进度条
        train_iterator.set_postfix(loss=loss.item(), mean_loss=mean_loss.item())
        all_predictions.extend(predicted.cpu().numpy())
        all_labels.extend(labels.cpu().numpy())

    train_accuracy = accuracy_score(y_true=all_labels, y_pred=all_predictions)
    return mean_loss.item(), train_accuracy


@torch.no_grad()
def eval_one_epoch(model, val_loader, device, criterion, epoch):
    result = {"y_pred": [], "y_true": [], 'val_loss': 0.}
    val_loss = torch.zeros(1).to(device)
    model.eval()
    val_iterator = tqdm(val_loader, file=sys.stdout, desc=f'{epoch + 1} epoch is validation...', colour='GREEN')
    for step, (images, labels) in enumerate(val_iterator):
        # 将数据转移到设备
        images, labels = images.to(device), labels.to(device)
        # 计算结果
        outputs = model(images)
        # 计算损失
        loss = criterion(outputs, labels)
        # 计算数据集上的全部损失
        val_loss = (val_loss * step + loss.detach()) / (step + 1)
        # 计算预测正确的样本
        _, predicted = torch.max(outputs, 1)
        val_iterator.set_postfix(loss=loss.item(), val_loss=val_loss.item())

        result["y_pred"].extend(predicted.cpu().numpy())
        result["y_true"].extend(labels.cpu().numpy())

    result['val_loss'] = val_loss.item()
    return result


def train_decae_one_epoch(model, train_loader, val_loader, device, optimizer, criterion, epoch):
    result = {'train_loss': 0., 'val_loss': 0., 'epoch': 0}
    train_loss = torch.zeros(1).to(device)
    model.train()
    train_iterator = tqdm(train_loader, file=sys.stdout, colour='yellow')
    for step, (images, label) in enumerate(train_iterator):
        images = images.to(device)

        optimizer.zero_grad()

        outputs = model(images)

        loss = criterion(outputs, images)

        loss.backward()
        optimizer.step()

        train_loss = (train_loss * step + loss.detach()) / (step + 1)
        train_iterator.set_postfix(loss=loss.item(), mean_loss=train_loss.item())

    print(f'the epoch {epoch + 1} train loss is {train_loss.item():.6f}')
    val_loss = torch.zeros(1).to(device)
    model.eval()
    val_iterator = tqdm(val_loader, file=sys.stdout, colour='MAGENTA')
    with torch.no_grad():
        for step, (images, label) in enumerate(val_iterator):
            images = images.to(device)

            outputs = model(images)
            loss = criterion(outputs, images)
            val_loss = (val_loss * step + loss.detach()) / (step + 1)

            val_iterator.set_postfix(loss=loss.item(), mean_loss=val_loss.item())
    print(f'the epoch {epoch + 1} val loss is {val_loss.item():.6f}')
    result['train_loss'] = train_loss.item()
    result['val_loss'] = val_loss.item()
    result['epoch'] = epoch + 1
    return result


def calculate_results(y_true, y_pred, classes):
    results = classification_report(y_true, y_pred, target_names=classes, output_dict=True, digits=4)
    return results
